# -*- coding: utf-8 -*-
"""
Boolean Network
================

Main class for Boolean network objects.

"""
#   Copyright (C) 2021 by
#   Rion Brattig Correia <rionbr@gmail.com>
#   Alex Gates <ajgates@indiana.edu>
#   Thomas Parmer <tjparmer@indiana.edu>
#   All rights reserved.
#   MIT license.
from collections import defaultdict
try:
    import cStringIO.StringIO as StringIO
except ImportError:
    from io import StringIO
import numpy as np
import networkx as nx
import random
import itertools
from cana.boolean_node import BooleanNode
import cana.bns as bns
from cana.control import fvs, mds, sc
from cana.utils import *
import warnings
import re
import copy


class BooleanNetwork:
    """

    """

    def __init__(self, name='', Nnodes=0, logic=None, sg=None, stg=None, stg_r=None, _ef=None, attractors=None,
                 constants={}, Nconstants=0, keep_constants=False,
                 bin2num=binstate_to_statenum, num2bin=statenum_to_binstate,
                 verbose=False, *args, **kwargs):

        self.name = name                            # Name of the Network
        self.Nnodes = Nnodes                        # Number of Nodes
        self.logic = logic                          # A dict that contains the network logic {<id>:{'name':<string>,'in':<list-input-node-id>,'out':<list-output-transitions>},..}
        self._sg = sg                               # Structure-Graph (SG)
        self._stg = stg                             # State-Transition-Graph (STG)
        self._stg_r = stg_r                         # State-Transition-Graph Reachability dict (STG-R)
        self._eg = None                             # Effective Graph, computed from the effective connectivity
        self._attractors = attractors               # Network Attractors
        #
        self.keep_constants = keep_constants        # Keep/Include constants in some of the computations
        self.constants = constants                  # A dict that contains of constant variables in the network
        self.Nconstants = len(constants)            # Number of constant variables
        #
        self.Nstates = 2**Nnodes                    # Number of possible states in the network 2^N
        #
        self.verbose = verbose

        # Intanciate BooleanNodes
        self.name2int = {logic[i]['name']: i for i in range(Nnodes)}
        self.Nself_loops = sum([self.name2int[logic[i]['name']] in logic[i]['in'] for i in range(Nnodes)])

        self.nodes = list()
        for i in range(Nnodes):
            name = logic[i]['name']
            k = len(logic[i]['in'])
            inputs = [self.name2int[logic[j]['name']] for j in logic[i]['in']]
            outputs = logic[i]['out']
            node = BooleanNode(id=i, name=name, k=k, inputs=inputs, outputs=outputs, network=self)
            self.nodes.append(node)

        self.input_nodes = [i for i in range(Nnodes) if (self.nodes[i].constant or ((self.nodes[i].k == 1) and (i in self.nodes[i].inputs)))]
        #
        self.bin2num = bin2num                      # Helper function. Converts binstate to statenum. It gets updated by `_update_trans_func`
        self.num2bin = num2bin                      # Helper function. Converts statenum to binstate. It gets updated by `_update_trans_func`
        self._update_trans_func()                   # Updates helper functions and other variables

    def __str__(self):
        node_names = [node.name for node in self.nodes]
        return "<BNetwork(name='{name:s}', N={number_of_nodes:d}, Nodes={nodes:})>".format(name=self.name, number_of_nodes=self.Nnodes, nodes=node_names)

    #
    # I/O Methods
    #
    @classmethod
    def from_file(self, file, type='cnet', keep_constants=True, **kwargs):
        """
        Load the Boolean Network from a file.

        Args:
            file (string) : The name of a file containing the Boolean Network.
            type (string) : The type of file, either 'cnet' (default) or 'logical' for Boolean logical rules.

        Returns:
            BooleanNetwork (object) : The boolean network object.

        See also:
            :func:`from_string` :func:`from_dict`
        """
        with open(file, 'r') as infile:
            if type == 'cnet':
                return self.from_string_cnet(infile.read(), keep_constants=keep_constants, **kwargs)
            elif type == 'logical':
                return self.from_string_boolean(infile.read(), keep_constants=keep_constants, **kwargs)

    @classmethod
    def from_string_cnet(self, string, keep_constants=True, **kwargs):
        """
        Instanciates a Boolean Network from a string in cnet format.

        Args:
            string (string): A cnet format representation of a Boolean Network.

        Returns:
            (BooleanNetwork)

        Examples:
            String should be structured as follow:

            .. code-block:: text

                #.v = number of nodes
                .v 1
                #.l = node label
                .l 1 node-a
                .l 2 node-b
                #.n = (node number) (in-degree) (input node 1) … (input node k)
                .n 1 2 4 5
                01 1 # transition rule

        See also:
            :func:`from_file` :func:`from_dict`
        """
        network_file = StringIO(string)
        logic = defaultdict(dict)

        line = network_file.readline()
        while line != "":
            if line[0] != '#' and line != '\n':
                # .v <#-nodes>
                if '.v' in line:
                    Nnodes = int(line.split()[1])
                    for inode in range(Nnodes):
                        logic[inode] = {'name': '', 'in': [], 'out': []}
                # .l <node-id> <node-name>
                elif '.l' in line:
                    logic[int(line.split()[1]) - 1]['name'] = line.split()[2]
                # .n <node-id> <#-inputs> <input-node-id>
                elif '.n' in line:
                    inode = int(line.split()[1]) - 1
                    indegree = int(line.split()[2])
                    for jnode in range(indegree):
                        logic[inode]['in'].append(int(line.split()[3 + jnode]) - 1)

                    logic[inode]['out'] = [0 for i in range(2**indegree) if indegree > 0]

                    logic_line = network_file.readline().strip()

                    if indegree <= 0:
                        if logic_line == '':
                            logic[inode]['in'] = [inode]
                            logic[inode]['out'] = [0, 1]
                        else:
                            logic[inode]['out'] = [int(logic_line)]
                    else:
                        while logic_line != '\n' and logic_line != '' and len(logic_line) > 1:
                            for nlogicline in expand_logic_line(logic_line):
                                logic[inode]['out'][binstate_to_statenum(nlogicline.split()[0])] = int(nlogicline.split()[1])
                            logic_line = network_file.readline().strip()

                # .e = end of file
                elif '.e' in line:
                    break
            line = network_file.readline()

        return self.from_dict(logic, keep_constants=keep_constants, **kwargs)

    @classmethod
    def from_string_boolean(self, string, keep_constants=True, **kwargs):
        """
        Instanciates a Boolean Network from a Boolean update rules format.

        Args:
            string (string) : A boolean update rules format representation of a Boolean Network.

        Returns:
            (BooleanNetwork) : The boolean network object.

        Examples:
            String should be structured as follow

            .. code-block:: text

                # BOOLEAN RULES (this is a comment)
                # node_name*=node_input_1 [logic operator] node_input_2 ...
                NODE3*=NODE1 AND NODE2 ...

        See also:
            :func:`from_string` :func:`from_dict`
        """

        logic = defaultdict(dict)

        # parse lines to receive node names
        network_file = StringIO(string)
        line = network_file.readline()
        i = 0
        while line != "":
            if line[0] == '#':
                line = network_file.readline()
                continue
            logic[i] = {'name': line.split("*")[0].strip(), 'in': [], 'out': []}
            line = network_file.readline()
            i += 1

        # Parse lines again to determine inputs and output sequence
        network_file = StringIO(string)
        line = network_file.readline()
        i = 0
        while line != "":
            if line[0] == '#':
                line = network_file.readline()
                continue
            eval_line = line.split("=")[1]  # logical condition to evaluate
            # RE checks for non-alphanumeric character before/after node name (node names are included in other node names)
            # Additional characters added to eval_line to avoid start/end of string complications
            input_names = [logic[node]['name'] for node in logic if re.compile('\W' + logic[node]['name'] + '\W').search('*' + eval_line + '*')]
            input_nums = [node for input in input_names for node in logic if input == logic[node]['name']]
            logic[i]['in'] = input_nums
            # Determine output transitions
            logic[i]['out'] = output_transitions(eval_line, input_names)
            line = network_file.readline()
            i += 1

        return self.from_dict(logic, keep_constants=keep_constants, **kwargs)

    @classmethod
    def from_dict(self, logic, keep_constants=True, **kwargs):
        """Instanciaets a BoolleanNetwork from a logic dictionary.

        Args:
            logic (dict) : The logic dict.
            keep_constants (bool) :

        Returns:
            (BooleanNetwork)

        See also:
            :func:`from_file` :func:`from_dict`
        """
        Nnodes = len(logic)
        keep_constants = keep_constants
        constants = {}
        if 'name' in kwargs:
            name = kwargs['name']
        else:
            name = ''
        if keep_constants:
            for i, nodelogic in logic.items():
                # No inputs? It's a constant!
                if len(nodelogic['in']) == 0:
                    constants[i] = logic[i]['out'][0]

        return BooleanNetwork(name=name, logic=logic, Nnodes=Nnodes, constants=constants, keep_constants=keep_constants)

    def to_cnet(self, file=None, adjust_no_input=False):
        """Outputs the network logic to ``.cnet`` format, which is similar to the Berkeley Logic Interchange Format (BLIF).
        This is the format used by BNS to compute attractors.

        Args:
            file (string,optional) : A string of the file to write the output to. If not supplied, a string will be returned.
            adjust_no_input (bool) : Adjust output string for nodes with no input.

        Returns:
            (string) : The ``.cnet`` format string.

        Note:
            See `BNS <https://people.kth.se/~dubrova/bns.html>`_ for more information.
        """
        # Copy
        logic = self.logic.copy()
        #
        if adjust_no_input:
            for i, data in logic.items():
                # updates in place
                if len(data['in']) == 0:
                    data['in'] = [i + 1]
                    data['out'] = [0, 1]

        bns_string = '.v ' + str(self.Nnodes) + '\n' + '\n'
        for i in range(self.Nnodes):
            k = len(logic[i]['in'])
            bns_string += '.n ' + str(i + 1) + " " + str(k) + " " + " ".join([str(v + 1) for v in logic[i]['in']]) + "\n"
            for statenum in range(2**k):
                # If is a constant (TODO: This must come from the BooleanNode, not the logic)
                if len(logic[i]['out']) == 1:
                    bns_string += str(logic[i]['out'][statenum]) + "\n"
                # Not a constant, print the state and output
                else:
                    bns_string += statenum_to_binstate(statenum, base=k) + " " + str(logic[i]['out'][statenum]) + "\n"
            bns_string += "\n"

        if file is None:
            return bns_string
        else:
            if isinstance(file, string):
                with open(file, 'w') as iofile:
                    iofile.write(bns_string)
                    iofile.close()
            else:
                raise AttributeError("File format not supported. Please specify a string.")

    #
    # Methods
    #
    def structural_graph(self, remove_constants=False):
        """Calculates and returns the structural graph of the boolean network.

        Args:
            remove_constants (bool) : Remove constants from the graph. Defaults to ``False``.

        Returns:
            G (networkx.Digraph) : The boolean network structural graph.
        """
        self._sg = nx.DiGraph(name="Structural Graph: " + self.name)

        # Add Nodes
        self._sg.add_nodes_from((i, {'label': n.name}) for i, n in enumerate(self.nodes))
        for target in range(self.Nnodes):
            for source in self.logic[target]['in']:
                self._sg.add_edge(source, target, **{'weight': 1.})

        if remove_constants:
            self._sg.remove_nodes_from(self.constants.keys())
        #
        return self._sg

    def number_interactions(self):
        """Returns the number of interactions in the Structural Graph (SG).
        Practically, it returns the number of edges of the SG.

        Returns:
            int
        """
        self._check_compute_variables(sg=True)
        return nx.number_of_edges(self._sg)

    def structural_indegrees(self):
        """Returns the in-degrees of the Structural Graph. Sorted.

        Returns:
            (int) : the number of in-degrees.

        See also:
            :func:`structural_outdegrees`, :func:`effective_indegrees`, :func:`effective_outdegrees`
        """
        self._check_compute_variables(sg=True)
        return sorted([d for n, d in self._sg.in_degree()], reverse=True)

    def structural_outdegrees(self):
        """Returns the out-degrees of the Structural Graph. Sorted.

        Returns:
            (list) : out-degrees.

        See also:
            :func:`structural_indegrees`, :func:`effective_indegrees`, :func:`effective_outdegrees`
        """
        self._check_compute_variables(sg=True)
        return sorted([d for n, d in self._sg.out_degree()], reverse=True)

    def effective_graph(self, bound='mean', threshold=None):
        """Computes and returns the effective graph of the network.
        In practive it asks each :class:`~cana.boolean_node.BooleanNode` for their :func:`~cana.boolean_node.BooleanNode.edge_effectiveness`.

        Args:
            bound (string) : The bound to which compute input redundancy.
                Can be one of : ["lower", "mean", "upper", "tuple"].
                Defaults to "mean".
            threshold (float) : Only return edges above a certain effective connectivity threshold.
                This is usefull when computing graph measures at diffent levels.

        Returns:
            (networkx.DiGraph) : directed graph

        See Also:
            :func:`~cana.boolean_node.BooleanNode.edge_effectiveness`
        """
        if threshold is not None:
            self._eg = nx.DiGraph(name="Effective Graph: " + self.name + "(Threshold: {threshold:.2f})".format(threshold=threshold))
        else:
            self._eg = nx.DiGraph(name="Effective Graph: " + self.name + "(Threshold: None)")

        # Add Nodes
        for i, node in enumerate(self.nodes, start=0):
            self._eg.add_node(i, **{'label': node.name})

        # Add Edges
        for i, node in enumerate(self.nodes, start=0):
            e_is = node.edge_effectiveness(bound=bound)
            for inputs, e_i in zip(self.logic[i]['in'], e_is):
                # If there is a threshold, only return those number above the threshold. Else, return all edges.
                if (threshold is None) or ((threshold is not None) and (e_i > threshold)):
                    self._eg.add_edge(inputs, i, **{'weight': e_i})

        return self._eg

    def conditional_effective_graph(self, conditioned_nodes={}, bound='mean', threshold=None):
        """Computes and returns the BN effective graph conditioned on some known states.

        Args:
            conditioned_nodes (dict) : a dictionary mapping node ids to their conditioned states.
                dict of form { nodeid : nodestate }
            bound (string) : The bound to which compute input redundancy.
                Can be one of : ["lower", "mean", "upper", "tuple"].
                Defaults to "mean".
            threshold (float) : Only return edges above a certain effective connectivity threshold.
                This is usefull when computing graph measures at diffent levels.

        Returns:
            (networkx.DiGraph) : directed graph

        See Also:
            :func:`~cana.boolean_network.BooleanNetwork.effective_graph`
        """

        conditional_eg = copy.deepcopy(self.effective_graph(bound=bound, threshold=None))
        conditioned_subgraph = set([])

        # make a copy of the logic dict so we can edit it
        conditioned_logic = copy.deepcopy(self.logic)

        # separate input conditioned nodes and nodes that get conditioned
        all_conditioned_nodes = dict(conditioned_nodes)
        # Queue of nodes to condition
        nodes2condition = list(conditioned_nodes.keys())

        while len(nodes2condition) > 0:

            conditioned_node = nodes2condition.pop(0)
            conditioned_value = str(all_conditioned_nodes[conditioned_node])
            conditioned_subgraph.add(conditioned_node)

            # take all successors of the conditioned node ignoring self-loops
            successors = [n for n in list(conditional_eg.neighbors(conditioned_node)) if n != conditioned_node]
            conditioned_subgraph.update(successors)

            # we have to loop through all of the successors of the conditioned node and change their logic
            for n in successors:

                # find the index of the conditioned node in the successor logic
                conditioned_node_idx = conditioned_logic[n]['in'].index(conditioned_node)

                conditioned_subgraph.update(conditioned_logic[n]['in'])

                # the new successor inputs without the conditioned node
                new_successor_inputs = conditioned_logic[n]['in'][:conditioned_node_idx] + conditioned_logic[n]['in'][(conditioned_node_idx + 1):]
                newk = len(new_successor_inputs)

                # now we create a conditioned LUT as the subset of the original for which the conditioned node is fixed to its value
                if newk == 0:
                    new_successor_outputs = [conditioned_logic[n]['out'][binstate_to_statenum(conditioned_value)]] * 2
                else:
                    new_successor_outputs = []
                    for sn in range(2**newk):
                        binstate = statenum_to_binstate(sn, newk)
                        binstate = binstate[:conditioned_node_idx] + conditioned_value + binstate[conditioned_node_idx:]
                        new_successor_outputs.append(conditioned_logic[n]['out'][binstate_to_statenum(binstate)])

                # use the new logic to calcuate a new edge effectiveness
                new_edge_effectiveness = BooleanNode().from_output_list(new_successor_outputs).edge_effectiveness(bound=bound)

                # and update the conditional effective graph with the new edge effectiveness values
                for i in range(newk):
                    conditional_eg[new_successor_inputs[i]][n]['weight'] = new_edge_effectiveness[i]

                # now update the conditioned_logic in case these nodes are further modified by additional conditioned variables
                conditioned_logic[n]['in'] = new_successor_inputs
                conditioned_logic[n]['out'] = new_successor_outputs

                # check if we just made a constant node
                if n not in all_conditioned_nodes and len(set(new_successor_outputs)) == 1:
                    # in which case, add it to the conditioned set and propagate the conditioned effect
                    nodes2condition.append(n)
                    all_conditioned_nodes[n] = new_successor_outputs[0]

        conditional_eg.name = "Conditioned Effective Graph: {name:s} conditioned on {nodes:s}".format(name=self.name, nodes=str(conditioned_nodes))
        if threshold is None:
            conditional_eg.name = conditional_eg.name + " (Threshold: None)"
        else:
            conditional_eg.name = conditional_eg.name + " (Threshold: {threshold:.2f})".format(threshold=threshold)
            remove_edges = [(i, j) for i, j, d in conditional_eg.edges(data=True) if d['weight'] <= threshold]
            conditional_eg.remove_edges_from(remove_edges)

        # add the conditional information into the effective graph object
        dict_conditioned_subgraph = {n: (n in conditioned_subgraph) for n in conditional_eg.nodes()}
        nx.set_node_attributes(conditional_eg, values=dict_conditioned_subgraph, name='conditioned_subgraph')

        dict_all_conditioned_nodes = {n: all_conditioned_nodes.get(n, None) for n in conditional_eg.nodes()}
        nx.set_node_attributes(conditional_eg, values=dict_all_conditioned_nodes, name='conditioned_state')

        return conditional_eg

    def effective_indegrees(self):
        """Returns the in-degrees of the Effective Graph. Sorted.

        Returns:
            (list)

        See also:
            :func:`effective_outdegrees`, :func:`structural_indegrees`, :func:`structural_outdegrees`
        """
        self._check_compute_variables(eg=True)
        return sorted([d for n, d in self._eg.in_degree()], reverse=True)

    def effective_outdegrees(self):
        """Returns the out-degrees of the Effective Graph. Sorted.

        Returns:
            (list)

        See also:
            :func:`effective_indegrees`, :func:`structural_indegrees`, :func:`structural_outdegrees`
        """
        self._check_compute_variables(eg=True)
        return sorted([d for n, d in self._eg.out_degree()], reverse=True)

    def activity_graph(self, threshold=None):
        """
        Returns the activity graph as proposed in
        Ghanbarnejad & Klemm (2012) EPL, 99

        Args:
            threshold (float) : Only return edges above a certain activity threshold.
                This is usefull when computing graph measures at diffent levels.

        Returns:
            (networkx.DiGraph) : directed graph

        """

        if threshold is not None:
            act_g = nx.DiGraph(name="Activity Graph: " + self.name + "(Threshold: %.2f)" % threshold)
        else:
            act_g = nx.DiGraph(name="Activity Graph: " + self.name + "(Threshold: None)")

        # Add Nodes
        for i, node in enumerate(self.nodes, start=0):
            act_g.add_node(i, **{'label': node.name})

        # Add Edges
        for i, node in enumerate(self.nodes, start=0):

            a_is = node.activities()
            for inputs, a_i in zip(self.logic[i]['in'], a_is):
                # If there is a threshold, only return those number above the threshold. Else, return all edges.
                if ((threshold is None) and (a_i > 0)) or ((threshold is not None) and (a_i > threshold)):
                    act_g.add_edge(inputs, i, **{'weight': a_i})

        return act_g

    def state_transition_graph(self):
        """Creates and returns the full State Transition Graph (STG) for the Boolean Network.

        Returns:
            (networkx.DiGraph) : The state transition graph for the Boolean Network.
        """
        self._stg = nx.DiGraph(name='STG: ' + self.name)
        self._stg.add_nodes_from((i, {'label': self.num2bin(i)}) for i in range(self.Nstates))
        for i in range(self.Nstates):
            b = self.num2bin(i)
            self._stg.add_edge(i, self.bin2num(self.step(b)))
        #
        return self._stg

    def stg_indegree(self):
        """Returns the In-degrees of the State-Transition-Graph (STG). Sorted.

        Returns:
            list
        """
        self._check_compute_variables(stg=True)
        return sorted(self._stg.in_degree().values(), reverse=True)

    def step(self, initial):
        """Steps the boolean network 'n' step from the given initial input condition.

        Args:
            initial (string) : the initial state.
            n (int) : the number of steps.

        Returns:
            (string) : The stepped binary state.
        """
        # for every node:
        #   node input = breaks down initial by node input
        #   asks node to step with the input
        #   append output to list
        # joins the results from each node output
        assert len(initial) == self.Nnodes
        return ''.join([node.step(node.input_mask(initial)) for node in self.nodes])

    def trajectory(self, initial, length=2):
        """Computes the trajectory of ``length`` steps without the State Transition Graph (STG)."""
        trajectory = [initial]
        for istep in range(length):
            trajectory.append(self.step(trajectory[-1]))
        return trajectory

    def trajectory_to_attractor(self, initial, precompute_attractors=True, return_attractor=False):
        """Computes the trajectory starting at `initial` until it reaches an attracor (this is garanteed).

        Args:
            initial (string) : the initial binstate.
            precompute_attractors (bool) : use precomputed attractors, default True.
            return_attractor (bool) : also return the attractor reached, default False.

        Returns:
            (list) : the state trajectory between initial and the final attractor state.
            if return_attractor: (list): the attractor
        """

        # if the attractors are already precomputed, then we can check when we reach a known state
        if precompute_attractors:
            self._check_compute_variables(attractors=True)
            attractor_states = [self.num2bin(s) for att in self._attractors for s in att]

            trajectory = [initial]
            while (trajectory[-1] not in attractor_states):
                trajectory.append(self.step(trajectory[-1]))

            if return_attractor:
                attractor = self.attractor(trajectory[-1])

        else:
            trajectory = [initial]
            while (trajectory[-1] not in trajectory[:-1]):
                trajectory.append(self.step(trajectory[-1]))

            # the attractor starts at the first occurence of the element
            idxatt = trajectory.index(trajectory[-1])
            if return_attractor:
                attractor = [self.bin2num(s) for s in trajectory[idxatt:-1]]
            trajectory = trajectory[:(idxatt + 1)]

        if return_attractor:
            return trajectory, attractor
        else:
            return trajectory

    def attractor(self, initial):
        """Computes the trajectory starting at ``initial`` until it reaches an attracor (this is garanteed)

        Args:
            initial (string): the initial state.
        Returns:
            attractor (string): the atractor state.
        """
        self._check_compute_variables(attractors=True)

        trajectory = self.trajectory_to_attractor(initial)
        for attractor in self._attractors:
            if self.bin2num(trajectory[-1]) in attractor:
                return attractor

    def attractors(self, mode='stg'):
        """Find the attractors of the boolean network.

        Args:
            mode (string) : ``stg`` or ``sat``. Defaults to ``stg``.
                ``stg``: Uses the full State Transition Graph (STG) and identifies the attractors as strongly connected components.
                ``bns``: Uses the SAT-based :mod:`cana.bns` to find all attractors.
        Returns:
            attractors (list) : A list containing all attractors for the boolean network.
        See also:
            :mod:`cana.bns`
        """
        self._check_compute_variables(stg=True)

        if mode == 'stg':
            self._attractors = [list(a) for a in nx.attracting_components(self._stg)]

        elif mode == 'bns':
            self._attractors = bns.attractors(self.to_cnet(file=None, adjust_no_input=False))
        else:
            raise AttributeError("Could not find the specified mode. Try 'stg' or 'bns'.")

        self._attractors.sort(key=len, reverse=True)
        return self._attractors

    def network_bias(self):
        """Network Bias. The sum of individual node biases divided by the number of nodes.
        Practically, it asks each node for their own bias.

        .. math:
            TODO

        See Also:
            :func:`~cana.boolean_node.BooleanNode.bias`
        """
        return sum([node.bias() for node in self.nodes]) / self.Nnodes

    def basin_entropy(self, base=2):
        """

        """
        self._check_compute_variables(stg=True)

        prob_vec = np.array([len(wcc) for wcc in nx.weakly_connected_components(self._stg)]) / 2.0**self.Nnodes
        return entropy(prob_vec, base=base)

    def set_constant(self, node, value=None):
        """Sets or unsets a node as a constant.

        Args:
            node (int) : The node ``id`` in the logic dict.

        Todo:
            This functions needs to better handle node_id and node_name
        """
        if value is not None:
            self.nodes[node].constant = True
            self.nodes[node].constant_value = value
            self.Nconstants += 1
        else:
            self.nodes[node].constant = False
            self.nodes[node].constant_value = value
            self.Nconstants -= 1

        self._update_trans_func()

    def remove_all_constants(self):
        self.keep_constants = False
        for inode in self.constants:
            self.set_constant(inode, None)

    def _update_trans_func(self):
        """

        """
        if self.keep_constants:
            self.Nstates = 2**(self.Nnodes - self.Nconstants)
            constant_template = [None if not (ivar in self.constants.keys()) else self.constants[ivar] for ivar in range(self.Nnodes)]
            self.bin2num = lambda bs: constantbinstate_to_statenum(bs, constant_template)
            self.num2bin = lambda sn: binstate_to_constantbinstate(
                statenum_to_binstate(sn, base=self.Nnodes - self.Nconstants), constant_template)
        else:
            self.Nstates = 2**self.Nnodes
            self.bin2num = binstate_to_statenum
            self.num2bin = lambda sn: statenum_to_binstate(sn, base=self.Nnodes)

    #
    # Dynamical Control Methods
    #
    def state_transition_graph_reachability(self, filename=None):
        """Generates a State-Transition-Graph Reachability (STG-R) dictionary.
        This dict/file will be used by the State Transition Graph Control Analysis.

        Args:
            filename (string) : The location to a file where the STG-R will be stored.

        Returns:
            (dict) : The STG-R in dict format.
        """
        self._check_compute_variables(stg=True)

        self._stg_r = {}

        if (filename is None):
            for source in self._stg:
                self._stg_r[source] = len(self._dfs_reachable(self._stg, source)) - 1.0
        else:
            try:
                with open(filename, 'rb') as handle:
                    self._stg_r = pickle.load(handle)
            except IOError:
                print("Finding STG dict")
                for source in self._stg:
                    self._stg_r[source] = len(self._dfs_reachable(self._stg, source)) - 1.0
                with open(filename, 'wb') as handle:
                    pickle.dump(self._stg_r, handle)
        return self._stg_r

    def attractor_driver_nodes(self, min_dvs=1, max_dvs=4, verbose=False):
        """Get the minimum necessary driver nodes by iterating the combination of all possible driver nodes of length :math:`min <= x <= max`.

        Args:
            min_dvs (int) : Mininum number of driver nodes to search.
            max_dvs (int) : Maximum number of driver nodes to search.

        Returns:
            (list) : The list of driver nodes found in the search.

        Note:
            This is an inefficient bruit force search, maybe we can think of better ways to do this?

        TODO:
            Parallelize the search on each combination. Each CSTG is independent and can be searched in parallel.

        See also:
            :func:`controlled_state_transition_graph`, :func:`controlled_attractor_graph`.
        """
        nodeids = list(range(self.Nnodes))
        if self.keep_constants:
            for cv in self.constants.keys():
                nodeids.remove(cv)

        attractor_controllers_found = []
        nr_dvs = min_dvs
        while (len(attractor_controllers_found) == 0) and (nr_dvs <= max_dvs):
            if verbose:
                print("Trying with {:d} Driver Nodes".format(nr_dvs))
            for dvs in itertools.combinations(nodeids, nr_dvs):
                dvs = list(dvs)
                # cstg = self.controlled_state_transition_graph(dvs)
                cag = self.controlled_attractor_graph(dvs)
                att_reachable_from = self.mean_reachable_attractors(cag)

                if att_reachable_from == 1.0:
                    attractor_controllers_found.append(dvs)
            # Add another driver node
            nr_dvs += 1

        if len(attractor_controllers_found) == 0:
            warnings.warn("No attractor control driver variable sets found after exploring all subsets of size {:,d} to {:,d} nodes!!".format(min_dvs, max_dvs))

        return attractor_controllers_found

    def controlled_state_transition_graph(self, driver_nodes=[]):
        """Returns the Controlled State-Transition-Graph (CSTG).
        In practice, it copies the original STG, flips driver nodes (variables), and updates the CSTG.

        Args:
            driver_nodes (list) : The list of driver nodes.

        Returns:
            (networkx.DiGraph) : The Controlled State-Transition-Graph.

        See also:
            :func:`attractor_driver_nodes`, :func:`controlled_attractor_graph`.
        """
        self._check_compute_variables(attractors=True)

        if self.keep_constants:
            for dv in driver_nodes:
                if dv in self.constants:
                    warnings.warn("Cannot control a constant variable '%s'! Skipping" % self.nodes[dv].name )

        # attractor_states = [s for att in self._attractors for s in att]
        cstg = copy.deepcopy(self._stg)
        cstg.name = 'C-' + cstg.name + ' (' + ','.join(map(str, [self.nodes[dv].name for dv in driver_nodes])) + ')'

        # add the control pertubations applied to all other configurations
        for statenum in range(self.Nstates):
            binstate = self.num2bin(statenum)
            controlled_states = flip_binstate_bit_set(binstate, copy.copy(driver_nodes))
            controlled_states.remove(binstate)

            for constate in controlled_states:
                cstg.add_edge(statenum, self.bin2num(constate))

        return cstg

    def pinning_controlled_state_transition_graph(self, driver_nodes=[]):
        """Returns a dictionary of Controlled State-Transition-Graph (CSTG)
        under the assumptions of pinning controllability.

        In practice, it copies the original STG, flips driver nodes (variables), and updates the CSTG.

        Args:
            driver_nodes (list) : The list of driver nodes.

        Returns:
            (networkx.DiGraph) : The Pinning Controlled State-Transition-Graph.

        See also:
            :func:`controlled_state_transition_graph`, :func:`attractor_driver_nodes`, :func:`controlled_attractor_graph`.
        """
        self._check_compute_variables(attractors=True)

        if self.keep_constants:
            for dv in driver_nodes:
                if dv in self.constants:
                    warnings.warn("Cannot control a constant variable {dv:s}'! Skipping".format(dv=self.nodes[dv].name))

        uncontrolled_system_size = self.Nnodes - len(driver_nodes)

        pcstg_dict = {}
        for att in self._attractors:
            dn_attractor_transitions = [tuple(''.join([self.num2bin(s)[dn] for dn in driver_nodes]) for s in att_edge)
                                        for att_edge in self._stg.subgraph(att).edges()]

            pcstg_states = [self.bin2num(binstate_pinned_to_binstate(
                            statenum_to_binstate(statenum, base=uncontrolled_system_size), attsource, pinned_var=driver_nodes))
                            for statenum in range(2**uncontrolled_system_size) for attsource, attsink in dn_attractor_transitions]

            pcstg = nx.DiGraph(name='STG: ' + self.name)
            pcstg.name = 'PC-' + pcstg.name + ' (' + ','.join(map(str, [self.nodes[dv].name for dv in driver_nodes])) + ')'

            pcstg.add_nodes_from((ps, {'label': ps}) for ps in pcstg_states)

            for attsource, attsink in dn_attractor_transitions:
                for statenum in range(2**uncontrolled_system_size):
                    initial = binstate_pinned_to_binstate(statenum_to_binstate(statenum, base=uncontrolled_system_size), attsource, pinned_var=driver_nodes)
                    pcstg.add_edge(self.bin2num(initial), self.bin2num(self.pinned_step(initial, pinned_binstate=attsink, pinned_var=driver_nodes)))

            pcstg_dict[tuple(att)] = pcstg

        return pcstg_dict

    def pinned_step(self, initial, pinned_binstate, pinned_var):
        """Steps the boolean network 1 step from the given initial input condition when the driver variables are pinned
        to their controlled states.

        Args:
            initial (string) : the initial state.
            n (int) : the number of steps.

        Returns:
            (string) : The stepped binary state.
        """
        # for every node:
        #   node input = breaks down initial by node input
        #   asks node to step with the input
        #   append output to list
        # joins the results from each node output
        assert len(initial) == self.Nnodes
        return ''.join([str(node.step(''.join(initial[j] for j in self.logic[i]['in']))) if not (i in pinned_var) else initial[i] for i, node in enumerate(self.nodes, start=0)])

    def controlled_attractor_graph(self, driver_nodes=[]):
        """
        Args:
            cstg (networkx.DiGraph) : A Controlled State-Transition-Graph (CSTG)

        Returns:
            (networkx.DiGraph) : The Controlled Attractor Graph (CAG)

        See also:
            :func:`attractor_driver_nodes`, :func:`controlled_state_transition_graph`.
        """
        self._check_compute_variables(attractors=True)

        if self.keep_constants:
            for dv in driver_nodes:
                if dv in self.constants:
                    warnings.warn("Cannot control a constant variable '%s'! Skipping" % self.nodes[dv].name )

        attractor_states = [s for att in self._attractors for s in att]
        cstg = copy.deepcopy(self._stg)
        cstg.name = 'C-' + cstg.name + ' Att(' + ','.join(map(str, [self.nodes[dv].name for dv in driver_nodes])) + ')'

        # add the control pertubations applied to only attractor configurations
        for statenum in attractor_states:
            binstate = self.num2bin(statenum)
            controlled_states = flip_binstate_bit_set(binstate, copy.copy(driver_nodes))
            controlled_states.remove(binstate)

            for constate in controlled_states:
                cstg.add_edge(statenum, self.bin2num(constate))

        Nattract = len(self._attractors)

        cag = nx.DiGraph(name='CAG: ' + cstg.name)
        # Nodes
        for i, attr in enumerate(self._attractors):
            cag.add_node(i, **{'label': '|'.join([self.num2bin(a) for a in attr])})
        # Edges
        for i in range(Nattract):
            ireach = self._dfs_reachable(cstg, self._attractors[i][0])
            for j in range(i + 1, Nattract):
                if self._attractors[j][0] in ireach:
                    cag.add_edge(i, j)
                if self._attractors[i][0] in self._dfs_reachable(cstg, self._attractors[j][0]):
                    cag.add_edge(j, i)
        return cag

    def mean_reachable_configurations(self, cstg):
        """Returns the Mean Fraction of Reachable Configurations

        Args:
            cstg (networkx.DiGraph) : The Controlled State-Transition-Graph.
        Returns:
            (float) : Mean Fraction of Reachable Configurations
        """
        reachable_from = []

        for source in cstg:
            control_reach = len(self._dfs_reachable(cstg, source)) - 1.0
            reachable_from.append(control_reach)

        norm = (2.0**self.Nnodes - 1.0) * len(reachable_from)
        reachable_from = sum(reachable_from) / (norm)

        return reachable_from

    def mean_controlable_configurations(self, cstg):
        """The Mean Fraction of Controlable Configurations

        Args:
            cstg (networkx.DiGraph) : The Controlled State-Transition-Graph.
        Returns:
            (float) : Mean Fraction of Controlable Configurations.
        """
        self._check_compute_variables(stg_r=True)

        control_from, reachable_from = [], []

        for source in cstg:
            control_reach = len(self._dfs_reachable(cstg, source)) - 1.0
            control_from.append(control_reach - self._stg_r[source])
            reachable_from.append(control_reach)

        norm = (2.0**self.Nnodes - 1.0) * len(reachable_from)
        control_from = sum(control_from) / (norm)

        return control_from

    def mean_reachable_attractors(self, cag, norm=True):
        """The Mean Fraction of Reachable Attractors to a specific Controlled Attractor Graph (CAG).

        Args:
            cag (networkx.DiGraph) : A Controlled Attractor Graph (CAG).

        Returns:
            (float) Mean Fraction of Reachable Attractors
        """
        att_norm = (float(len(cag)) - 1.0) * len(cag)

        if att_norm == 0:
            # if there is only one attractor everything is reachable
            att_reachable_from = 1
        else:
            # otherwise find the reachable from each attractor
            att_reachable_from = [len(self._dfs_reachable(cag, idxatt)) - 1.0 for idxatt in cag]
            att_reachable_from = sum(att_reachable_from) / (att_norm)

        return att_reachable_from

    def fraction_pinned_attractors(self, pcstg_dict):
        """Returns the Number of Accessible Attractors

        Args:
            pcstg_dict (dict of networkx.DiGraph) : The dictionary of Pinned Controlled State-Transition-Graphs.

        Returns:
            (int) : Number of Accessible Attractors
        """
        reached_attractors = []
        for att, pcstg in pcstg_dict.items():
            pinned_att = list(nx.attracting_components(pcstg))
            print(set(att), pinned_att)
            reached_attractors.append(set(att) in pinned_att)
        return sum(reached_attractors) / float(len(pcstg_dict))

    def fraction_pinned_configurations(self, pcstg_dict):
        """Returns the Fraction of successfully Pinned Configurations

        Args:
            pcstg_dict (dict of networkx.DiGraph) : The dictionary of Pinned Controlled State-Transition-Graphs.

        Returns:
            (list) : the Fraction of successfully Pinned Configurations to each attractor
        """
        pinned_configurations = []
        for att, pcstg in pcstg_dict.items():
            att_reached = False
            for wcc in nx.weakly_connected_components(pcstg):
                if set(att) in list(nx.attracting_components(pcstg.subgraph(wcc))):
                    pinned_configurations.append(len(wcc) / len(pcstg))
                    att_reached = True
            if not att_reached:
                pinned_configurations.append(0)

        return pinned_configurations

    def mean_fraction_pinned_configurations(self, pcstg_dict):
        """Returns the mean Fraction of successfully Pinned Configurations

        Args:
            pcstg_dict (dict of networkx.DiGraph) : The dictionary of Pinned Controlled State-Transition-Graphs.

        Returns:
            (int) : the mean Fraction of successfully Pinned Configurations
        """
        return sum(self.fraction_pinned_configurations(pcstg_dict)) / len(pcstg_dict)

    def _dfs_reachable(self, G, source):
        """Produce nodes in a depth-first-search pre-ordering starting from source."""
        return [n for n in nx.dfs_preorder_nodes(G, source)]

    #
    # Feedback Vertex Set (FVS)
    #
    def feedback_vertex_set_driver_nodes(self, graph='structural', method='grasp', max_iter=1, max_search=11, keep_self_loops=True, *args, **kwargs):
        """The minimum set of necessary driver nodes to control the network based on Feedback Vertex Set (FVS) theory.

        Args:
            graph (string) : Which graph to perform computation
            method (string) : FVS method. ``bruteforce`` or ``grasp`` (default).
            max_iter (int) : The maximum number of iterations used by the grasp method.
            max_search (int) : The maximum number of searched used by the bruteforce method.
            keep_self_loops (bool) : Keep or remove self loop in the graph to be searched.

        Returns:
            (list) : A list-of-lists with FVS solution nodes.

        Note:
            When computing FVS on the structural graph, you might want to use ``remove_constants=True``
            to make sure the resulting set is minimal – since constants are not controlabled by definition.
            Also, when computing on the effective graph, you can define the desired ``threshold`` level.
        """
        self._check_compute_variables(sg=True)

        if graph == 'structural':
            dg = self.structural_graph(*args, **kwargs)
        elif graph == 'effective':
            dg = self.effective_graph(mode='input', bound='mean', threshold=None, *args, **kwargs)
        else:
            raise AttributeError("The graph type '%s' is not accepted. Try 'structural' or 'effective'." % graph)
        #
        if method == 'grasp':
            fvssets = fvs.fvs_grasp(dg, max_iter=max_iter, keep_self_loops=keep_self_loops)
        elif method == 'bruteforce':
            fvssets = fvs.fvs_bruteforce(dg, max_search=max_search, keep_self_loops=keep_self_loops)
        else:
            raise AttributeError("The FVS method '%s' does not exist. Try 'grasp' or 'bruteforce'." % method)

        fvssets = [fvc.union(set(self.input_nodes)) for fvc in fvssets]

        return fvssets  # [ [self.nodes[i].name for i in fvsset] for fvsset in fvssets]

    #
    # Minimum Dominating Set
    #
    def minimum_dominating_set_driver_nodes(self, graph='structural', max_search=5, keep_self_loops=True, *args, **kwargs):
        """The minimun set of necessary driver nodes to control the network based on Minimum Dominating Set (MDS) theory.

        Args:
            max_search (int) : Maximum search of additional variables. Defaults to 5.
            keep_self_loops (bool) : If self-loops are used in the computation.

        Returns:
            (list) : A list-of-lists with MDS solution nodes.
        """
        self._check_compute_variables(sg=True)
        #
        if graph == 'structural':
            dg = self.structural_graph(*args, **kwargs)
        elif graph == 'effective':
            dg = self.effective_graph(mode='input', bound='mean', threshold=None, *args, **kwargs)
        else:
            raise AttributeError("The graph type '%s' is not accepted. Try 'structural' or 'effective'." % graph)
        #

        mdssets = mds.mds(dg, max_search=max_search, keep_self_loops=keep_self_loops)
        return mdssets  # [ [self.nodes[i].name for i in mdsset] for mdsset in mdssets]

    # Structural Controllability
    #
    def structural_controllability_driver_nodes(self, graph='structural', keep_self_loops=True, *args, **kwargs):
        """The minimum set of necessary driver nodes to control the network based on Structural Controlability (SC) theory.

        Args:
            keep_self_loops (bool) : If self-loops are used in the computation.

        Returns:
            (list) : A list-of-lists with SC solution nodes.
        """
        self._check_compute_variables(sg=True)

        if graph == 'structural':
            dg = self.structural_graph(*args, **kwargs)
        elif graph == 'effective':
            dg = self.effective_graph(mode='input', bound='mean', threshold=None, *args, **kwargs)
        else:
            raise AttributeError("The graph type '%s' is not accepted. Try 'structural' or 'effective'." % graph)
        #
        scsets = [set(scset).union(set(self.input_nodes)) for scset in sc.sc(dg, keep_self_loops=keep_self_loops)]
        return scsets  # [ [self.nodes[i].name for i in scset] for scset in scsets]

    #
    # Dynamical Impact
    #
    def partial_derative_node(self, node, n_traj=10, t=1):
        """The partial derivative of node on all other nodes after t steps

        Args:
            node (int) : the node index for perturbations

            t (int) : the number of time steps the system is run before impact is calculated.

            n_traj (int) : the number of trajectories used to approximate the dynamical impact of a node.
                if 0 then the full STG is used to calculate the true value instead of the approximation method.

        Returns:
            (vector) : the partial derivatives
        """
        partial = np.zeros((t, self.Nnodes), dtype=float)
        if n_traj == 0:
            config_genderator = (self.num2bin(statenum) for statenum in range(self.Nstates))
            n_traj = self.Nstates
        else:
            # sample configurations
            config_genderator = (random_binstate(self.Nnodes) for itraj in range(n_traj))

        for config in config_genderator:
            perturbed_config = flip_binstate_bit(config, node)
            for n_step in range(t):
                config = self.step(config)
                perturbed_config = self.step(perturbed_config)
                partial[n_step] += np.logical_not(binstate_compare(config, perturbed_config))
        partial /= n_traj

        return partial

    def approx_dynamic_impact(self, source, n_steps=1, target_set=None, bound='mean', threshold=0.0):
        """Use the network structure to approximate the dynamical impact of a perturbation to node for each of n_steps
        for details see: Gates et al (2020).

        Args:
            source (int) : the source index for perturbations
            n_steps (int) : the number of time steps

        bound (str) : the bound for the effective graph
            'mean' - edge effectiveness
            'upper' - activity

        Returns:
            (matrix) : approximate dynamical impact for each node at each step (2 x n_steps x n_nodes)
        """

        if target_set is None:
            target_set = range(self.Nnodes)

        Gstr = self.structural_graph()

        Geff = self.effective_graph(bound=bound, threshold=threshold)

        # the maximum path with length given by product of weights is the same as minimal path of negative log weight
        def eff_weight_func(u, v, e):
            return -np.log(e['weight'])

        def inv_eff_weight_func(pathlength):
            return np.exp(-pathlength)


        impact_matrix = np.zeros((2, n_steps + 1, len(target_set)))
        impact_matrix[0, :, :] = self.Nnodes + 1  # if we can't reach the node, then the paths cant be longer than the number of nodes in the graph
        # note that by default: impact_matrix[1, :, :] = 0 the minimum path for nodes we cant reach in the effective graph

        # in the structural graph, calcluate the dijkstra shortest paths from the source to all targets that are shorter than the cufoff
        Gstr_shortest_dist, Gstr_shortest_paths = nx.single_source_dijkstra(Gstr, source, target=None, cutoff=n_steps)
        Gstr_shortest_dist = {n: int(l) for n, l in Gstr_shortest_dist.items()}

        # in the effective graph, calcluate the dijkstra shortest paths from the source to all targets that are shorter than the cufoff
        # where the edge weight is given by the effective weight function
        Geff_shortest_dist, Geff_shortest_paths = nx.single_source_dijkstra(Geff, source, target=None, cutoff=n_steps, weight=eff_weight_func)

        for itar, target in enumerate(target_set):

            # we dont need to worry about a path to iteself (source==target)
            # and if the target doesnt appear in the shortest path dict, then no path exists that is less than the cutoff
            if target != source and not Gstr_shortest_dist.get(target, None) is None:

                # the light cone is at least as big as the number of edges in the structural shorest path
                impact_matrix[0, list(range(Gstr_shortest_dist[target], n_steps + 1)), itar] = Gstr_shortest_dist[target]

                # if the path exists, then the number of edges (timesteps) is one less than the number of nodes
                if not Geff_shortest_paths.get(target, None) is None:
                    eff_path_steps = len(Geff_shortest_paths[target]) - 1
                else:
                    # or the path doesnt exist
                    eff_path_steps = n_steps + 100 # any number bigger than the longest path to represent we cannot reach the node


                # start by checking if the number of timesteps is less than the maximum allowable number of steps
                if eff_path_steps <= n_steps:

                    # now check if the most likely effective path is longer (in terms of # of timesteps) than the structural shortest path
                    if eff_path_steps > Gstr_shortest_dist[target]:

                        # if it is, then we need to find another effective path constrained by the light-cone
                        # for all time steps where the most likely effective path is longer (in terms of # of timesteps)
                        # than the structural shortest path
                        for istep in range(Gstr_shortest_dist[target], eff_path_steps):

                            # bc the effective graph has fully redundant edges, there may actually not be a path
                            try:
                                redo_dijkstra_dist, _ = nx.single_source_dijkstra(Geff,
                                    source=source,
                                    target=target,
                                    cutoff=istep,
                                    weight=eff_weight_func)
                                impact_matrix[1, istep, itar] = inv_eff_weight_func(redo_dijkstra_dist)
                            except nx.NetworkXNoPath:
                                pass

                    # once the lightcone includes the target node on the effective shortest path,
                    # then for all other steps the effective path is the best
                    impact_matrix[1, list(range(eff_path_steps, n_steps + 1)), itar] = inv_eff_weight_func(Geff_shortest_dist[target])

        return impact_matrix[:, 1:]


    def dist_from_attractor(self):
        """Find the distance from attractor for each configuration.

        Returns:
            distance (dict). Nodes are dictionary indexes and distances the values.
        """
        self._check_compute_variables(attractors=True)

        dist = {}  # stores [node, distance] pair
        for att in self._attractors:
            dist.update({a: (0, a) for a in att})
            dag = nx.bfs_tree(self._stg, att[0], reverse=True)
            attractor_states = set(att)
            for node in nx.topological_sort(dag):
                # pairs of dist,node for all incoming edges
                if node not in attractor_states:
                    pairs = [(dist[v][0] + 1, v) for v in dag.pred[node]]
                    if pairs:
                        dist[node] = min(pairs)
                    else:
                        dist[node] = (0, node)

        return dist

    def average_dist_from_attractor(self):
        dist = self.dist_from_attractor()
        return np.mean([d[0] for d in dist.values() if d[0] > 0])

    #
    # Dynamics Canalization Map (DCM)
    #
    def dynamics_canalization_map(self, output=None, simplify=True):
        """Computes the Dynamics Canalization Map (DCM).
        In practice, it asks each node to compute their Canalization Map and then puts them together, simplifying it if possible.

        Args:
            output (int) : The output DCM to return. Default is ``None``, retuning both [0,1].
            simplify (bool) : Attemps to simpify the DCM by removing thresholds nodes with :math:`\tao=1`.

        Returns:
            DCM (networkx.DiGraph) : a directed graph representation of the DCM.

        See Also:
            :func:`boolean_node.canalizing_map` for the CM and :func:`drawing.draw_dynamics_canalizing_map_graphviz` for plotting.
        """
        CMs = []
        for node in self.nodes:
            if self.keep_constants or not node.constant:
                CMs.append(node.canalizing_map(output))
        # https://networkx.github.io/documentation/stable/reference/algorithms/generated/networkx.algorithms.operators.all.compose_all.html
        DCM = nx.compose_all(CMs)
        DCM.name = 'DCM: %s' % (self.name)

        if simplify:
            # Loop all threshold nodes
            threshold_nodes = [(n, d) for n, d in DCM.nodes(data=True) if d['type'] == 'threshold']
            for n, d in threshold_nodes:

                # Constant, remove threshold node
                if d['tau'] == 0:
                    DCM.remove_node(n)

                # Tau == 1
                if d['tau'] == 1:
                    in_nei = list(DCM.in_edges(n))[0]
                    out_nei = list(DCM.out_edges(n))[0]

                    # neis = set(list(in_nei) + list(out_nei))

                    # Convert to self loop
                    if (in_nei == out_nei[::-1]):
                        DCM.remove_node(n)
                        DCM.add_edge(in_nei[0], out_nei[1], **{'type': 'simplified', 'mode': 'selfloop'})
                    # Link variables nodes directly
                    elif not any([DCM.nodes[tn]['type'] == 'fusion' for tn in in_nei]):
                        DCM.remove_node(n)
                        DCM.add_edge(in_nei[0], out_nei[1], **{'type': 'simplified', 'mode': 'direct'})
        # Remove Isolates
        isolates = list(nx.isolates(DCM))
        DCM.remove_nodes_from(isolates)

        return DCM

    def _check_compute_variables(self, **kwargs):
        """Recursevely check if the requested control variables are instantiated/computed.
        Otherwise computes them in order.
        """
        if 'sg' in kwargs:
            if self._sg is None:
                self._sg = self.structural_graph()

        elif 'eg' in kwargs:
            if self._eg is None:
                self._eg = self.effective_graph()

        elif 'stg' in kwargs:
            if self._stg is None:
                self._check_compute_variables(sg=True)
                self._stg = self.state_transition_graph()

        elif 'attractors' in kwargs:
            if self._attractors is None:
                self._check_compute_variables(stg=True)
                self._attractors = self.attractors()

        elif 'stg_r' in kwargs:
            if self._stg_r is None:
                self._check_compute_variables(stg=True)
                self._stg_r = self.state_transition_graph_reachability()
        else:
            raise Exception('Control variable name not found. %s' % kwargs)
        return True

    #
    # Get Node Names from Ids
    #
    def _get_node_name(self, id):
        """Return the name of the node based on its id.

        Args:
            id (int): id of the node.

        Returns:
            name (string): name of the node.
        """
        try:
            node = self.nodes[id]
        except error:
            raise AttributeError("Node with id {id:d} does not exist. {error::s}".format(id=id, error=error))
        else:
            return node.name

    def get_node_name(self, iterable=[]):
        """Return node names. Optionally, it returns only the names of the ids requested.

        Args:
            iterable (int,list, optional) : The id (or list of ids) of nodes to which return their names.

        Returns:
            names (list) : The name of the nodes.
        """
        # If only one id is passed, make it a list
        if not isinstance(iterable, list):
            iterable = [iterable]
        # No ids requested, return all the names
        if not len(iterable):
            return [n.name for n in self.nodes]
        # otherwise, use the recursive map to change ids to names
        else:
            return recursive_map(self._get_node_name, iterable)

    def average_trajectory_length(self, nsamples=10, random_seed=None, method='random'):
        """The average length of trajectories from a random initial configuration to its attractor.

        Args:
            nsamples (int) : The number of samples per hammimg distance to get.
            random_seed (int) : The random state seed.
            method (string) : specify the method you want. either 'random' or ....

        Returns:
            trajlen (float) : The average trajectory length to an attractor.
        """
        return sum(len(self.trajectory_to_attractor(random_binstate(self.Nnodes))) for isample in range(nsamples)) / nsamples

    def derrida_curve(self, nsamples=10, max_hamm=None, random_seed=None, method='random'):
        """The Derrida Curve (also reffered as criticality measure :math:`D_s`).
        When "mode" is set as "random" (default), it would use random sampling to estimate Derrida value
        If "mode" is set as "sensitivity", it would use c-sensitivity to calculate Derrida value (slower)
        You can refer to :cite:'kadelka2017influence' about why c-sensitivity can be used to caculate Derrida value

        Args:
            nsamples (int) : The number of samples per hammimg distance to get.
            max_hamm (int) : The maximum Hamming distance between starting states. default: self.Nnodes
            random_seed (int) : The random state seed.
            method (string) : specify the method you want. either 'random' or 'sensitivity'

        Returns:
            (dx,dy) (tuple) : The dx and dy of the curve.
        """
        random.seed(random_seed)

        if max_hamm is None or (max_hamm > self.Nnodes):
            max_hamm = self.Nnodes

        dx = np.linspace(0, 1, max_hamm, endpoint=True)
        dy = np.zeros(max_hamm + 1)

        if method == 'random':
            # for each possible hamming distance between the starting states
            for hamm_dist in range(1, max_hamm + 1):

                # sample nsample times
                for isample in range(nsamples):
                    rnd_config = random_binstate(self.Nnodes)
                    perturbed_var = random.sample(range(self.Nnodes), hamm_dist)
                    perturbed_config = [flip_bit(rnd_config[ivar]) if ivar in perturbed_var else rnd_config[ivar] for ivar in range(self.Nnodes)]
                    dy[hamm_dist] += hamming_distance(self.step(rnd_config), self.step(perturbed_config)) / self.Nnodes  # normalized Hamming Distance

            dy /= nsamples

        elif method == 'sensitivity':

            for hamm_dist in range(1, max_hamm + 1):
                dy[hamm_dist] = sum([node.c_sensitivity(hamm_dist, mode='forceK', max_k=self.Nnodes) for node in self.nodes]) / self.Nnodes

        return dx, dy

    def derrida_coefficient(self, nsamples=10, random_seed=None, method='random'):
        """The Derrida Coefficient.
        When "mode" is set as "random" (default), it would use random sampling to estimate Derrida value
        If "mode" is set as "sensitivity", it would use c-sensitivity to calculate Derrida value (slower)
        You can refer to :cite:'kadelka2017influence' about why c-sensitivity can be used to caculate Derrida value

        Args:
            nsamples (int) : The number of samples per hammimg distance to get.
            random_seed (int) : The random state seed.
            method (string) : specify the method you want. either 'random' or 'sensitivity'

        Returns:
            (dx,dy) (tuple) : The dx and dy of the curve.
        """
        random.seed(random_seed)
        hamm_dist = 1

        if method == 'random':
            # for each possible hamming distance between the starting states

            dy = 0
            # sample nsample times
            for isample in range(nsamples):
                rnd_config = random_binstate(self.Nnodes)
                perturbed_var = random.sample(range(self.Nnodes), hamm_dist)
                perturbed_config = [flip_bit(rnd_config[ivar]) if ivar in perturbed_var else rnd_config[ivar] for ivar in range(self.Nnodes)]
                dy += hamming_distance(self.step(rnd_config), self.step(perturbed_config))

            dy /= float(nsamples)

        elif method == 'sensitivity':
            # raise NotImplementedError
            dy = sum([node.c_sensitivity(hamm_dist, mode='forceK', max_k=self.Nnodes) for node in self.nodes])

        return dy / float(self.Nnodes)



import networkx as nx
import numpy as np
from itertools import takewhile
import copy
import math
import operator as op
from functools import reduce
from cana.cutils import *


def flip_bitset_in_strstates(strstates, idxs):
    """Flips the binary value for a set of bits in a binary state.

    Args:
        binstate (string) : The binary state to flip.
        idxs (int) : The indexes of the bits to flip.

    Returns:
        (list) : The flipped states

    Example:
        >>> flip_bit_in_strstates('000',[0, 2])
        ['100','001']
    """
    return [flip_bit_in_strstates(strstates, idx) for idx in idxs]


def flip_binstate_bit_set(binstate, idxs):
    """Flips the binary value for a set of bits in a binary state.

    Args:
        binstate (string) : The binary state to flip.
        idxs (int) : The indexes of the bits to flip.

    Returns:
        (list) : The flipped states
    """
    flipset = []
    if (len(idxs) != 0):
        fb = idxs.pop()
        flipset.extend(flip_binstate_bit_set(binstate, copy.copy(idxs)))
        flipset.extend(flip_binstate_bit_set(flip_binstate_bit(binstate, fb), copy.copy(idxs)))
    else:
        flipset.append(binstate)
    return flipset


def print_logic_table(outputs):
    """Print Logic Table

    Args:
        outputs (list) : The transition outputs of the function.

    Returns:
        print : a print-out of the logic table.

    Example:
        >>> print_logic_table([0,0,1,1])
        00 : 0
        01 : 0
        10 : 1
        11 : 1

    """
    k = int(math.log(len(outputs)) / math.log(2))
    for statenum in range(2**k):
        print(statenum_to_binstate(statenum, base=k) + " : " + str(outputs[statenum]))


def entropy(prob_vector, logbase=2.):
    """Calculates the entropy given a probability vector

    Todo:
        This should be calculated using ``scipy.entropy``
    """
    prob_vector = np.array(prob_vector)
    pos_prob_vector = prob_vector[prob_vector > 0]
    return - np.sum(pos_prob_vector * np.log(pos_prob_vector) / np.log(logbase))


def ncr(n, r):
    """Return the combination number.
    The combination of selecting `r` items from `n` iterms, order doesn't matter.

    Args:
        n (int): number of elements in collection
        r (int): length of combination

    Returns:
        int
    """
    r = min(r, n - r)
    if r == 0:
        return 1
    numer = reduce(op.mul, range(n, n - r, -1))
    denom = reduce(op.mul, range(1, r + 1))
    return numer // denom


def isclose(a, b, rel_tol=1e-09, abs_tol=0.0):
    """Python 2 doesn't have math.isclose()
    Here is an equivalent function
    Use this to tell whether two float numbers are close enough
    considering using == to compare floats is dangerous!
    2.0*3.3 != 3.0*2.2 in python!

    Args:
        a (float) : the first float number
        b (float) : the second float number
        rel_tol (float) : the relative difference threshold between a and b
        abs_tol (float) : absolute difference threshold. not recommended for float

    Returns:
        bool
    """
    return abs(a - b) <= max(rel_tol * max(abs(a), abs(b)), abs_tol)


def output_transitions(eval_line, input_list):
    """Returns an output list from combinatorically trying all input values

    Args:
        eval_line (string) : logic or arithmetic line to evaluate
        input_list (list) : list of input variables

    Returns:
        list of all possible output transitions (list)

    Example:
        RAS*=(GRB2 or PLCG1) and not GAP

        .. code-block:: python

            >>> eval_line = "(GRB2 or PLCG1) and not GAP"
            >>> input_list = ['GRB2', 'PLCG1', 'GAP']
            >>> output_transitions(eval_line, input_list)
            000
            001
            010
            011
            100
            101
            110
            111

        A variable is dynamically created for each member of the input list
        and assigned the corresponding value from each trail string.
        The original eval_line is then evaluated with each assignment
        which results in the output list [0, 0, 1, 0, 1, 0, 1, 0]
    """
    total = 2**len(input_list)  # Total combinations to try
    output_list = []
    for i in range(total):
        trial_string = statenum_to_binstate(i, len(input_list))
        # Evaluate trial_string by assigning value to each input variable
        for j, input in enumerate(input_list):
            exec(input + '=' + trial_string[j])
        output_list.append(int(eval(eval_line)))

    return output_list


def seq_upto(seq, obj):
    """
    TODO: description
    """
    return takewhile(lambda el: el != obj, iter(seq))


def mindist_from_source(G, source):
    """
    TODO: description
    """
    dag = nx.bfs_tree(G, source)
    dist = {}  # stores [node, distance] pair
    for node in nx.topological_sort(dag):
        # pairs of dist,node for all incoming edges
        pairs = [(dist[v][0] + 1, v) for v in dag.pred[node]]
        if pairs:
            dist[node] = min(pairs)
        else:
            dist[node] = (0, node)

    return dist


def pathlength(p, weights, rule='sum'):
    """Calculate the length of path p, with weighted edges, given the length rule of:

    Ars:
        weights:

        rule (str):
            'sum' - sum of edge weights along path
            'prod' - product of edge weights along path
            'min' - minimum of edge weights along path (weakest-link)
            'max' - maximum of edge weights along path

    TODO: update description
    """
    if rule == 'sum':
        return sum(weights[(p[ie], p[ie + 1])] for ie in range(len(p) - 1))
    elif rule == 'prod':
        return np.prod([weights[(p[ie], p[ie + 1])] for ie in range(len(p) - 1)])
    elif rule == 'min':
        return min(weights[(p[ie], p[ie + 1])] for ie in range(len(p) - 1))
    elif rule == 'max':
        return max(weights[(p[ie], p[ie + 1])] for ie in range(len(p) - 1))



# -*- coding: utf-8 -*-
"""
Boolean Nodes
=================================

Commonly used boolean node functions.

"""
#   Copyright (C) 2021 by
#   Alex Gates <ajgates42@gmail.com>
#   Rion Brattig Correia <rionbr@gmail.com>
#   All rights reserved.
#   MIT license.
from .. boolean_network import BooleanNode


def AND():
    """AND boolean node.

    .. code::

        00 : 0
        01 : 0
        10 : 0
        11 : 1

    """
    return BooleanNode.from_output_list(outputs=[0, 0, 0, 1], name="AND")


def OR():
    """OR boolean node.

    .. code::

        00 : 0
        01 : 1
        10 : 1
        11 : 1

    """
    return BooleanNode.from_output_list(outputs=[0, 1, 1, 1], name="OR")


def XOR():
    """XOR boolean node.

    .. code::

        00 : 0
        01 : 1
        10 : 1
        11 : 0

    """
    return BooleanNode.from_output_list(outputs=[0, 1, 1, 0], name="XOR")


def COPYx1():
    """COPY :math:`x_1` boolean node.

    .. code::

        00 : 0
        01 : 0
        10 : 1
        11 : 1

    """
    return BooleanNode.from_output_list(outputs=[0, 0, 1, 1], name="COPY x_1")


def CONTRADICTION():
    """Contradiction boolean node.

    .. code::

        00 : 0
        01 : 0
        10 : 0
        11 : 0

    """
    return BooleanNode.from_output_list(outputs=[0, 0, 0, 0], name="CONTRADICTION")


def RULE90():
    """RULE 90 celular automata node.

    .. code::

        000 : 0
        001 : 1
        010 : 0
        011 : 1
        100 : 1
        101 : 0
        110 : 1
        111 : 0

    """
    return BooleanNode.from_output_list(outputs=[0, 1, 0, 1, 1, 0, 1, 0], name="RULE 90")


def RULE110():
    """RULE 110 celular automata node.

    .. code::

        000 : 0
        001 : 1
        010 : 1
        011 : 1
        100 : 0
        101 : 1
        110 : 1
        111 : 0

    """
    return BooleanNode.from_output_list(outputs=[0, 1, 1, 1, 0, 1, 1, 0], name="RULE 110")



import warnings
import functools


def deprecated(func):
    """This is a decorator which can be used to mark functions as deprecated.
    It will result in a warning being emitted when the function is used.
    """
    @functools.wraps(func)
    def new_func(*args, **kwargs):
        warnings.warn_explicit(
            "You've called a deprecated function. Maybe this function needs updating to the new package. {}".format(func.__name__),
            category=DeprecationWarning,
            filename=func.func_code.co_filename,
            lineno=func.func_code.co_firstlineno + 1
        )
        return func(*args, **kwargs)
    return new_func



# -*- coding: utf-8 -*-
"""
Random Boolean Network
======================

Methods to generate random ensembles of Boolean networks.

"""
#   Copyright (C) 2021 by
#   Alex Gates <ajgates@indiana.edu>
#   Rion Brattig Correia <rionbr@gmail.com>
#   Thomas Parmer <tjparmer@indiana.edu>
#   All rights reserved.
#   MIT license.
from collections import defaultdict, Counter
import networkx as nx
import random
from cana.boolean_network import BooleanNetwork
import re
from io import StringIO
from cana.utils import output_transitions


def regular_boolean_network(N=10, K=2, bias=0.5, bias_constraint='soft', keep_constants=True, remove_multiedges=True, niter_remove=1000):
    """
    TODO: description
    """
    din = [K] * N  # in-degree distrubtion
    dout = [K] * N  # out-degree distrubtion

    regular_graph = nx.directed_configuration_model(din, dout)

    # the configuration graph creates a multigraph with self loops
    # the self loops are OK, but we should only have one copy of each edge
    if remove_multiedges:
        regular_graph = _remove_duplicate_edges(graph=regular_graph, niter_remove=niter_remove)

    # A dict that contains the network logic {<id>:{'name':<string>,'in':<list-input-node-id>,'out':<list-output-transitions>},..}
    bn_dict = {
        node: {
            'name': str(node),
            'in': sorted([n for n in regular_graph.predecessors(node)]),
            'out': random_automata_table(regular_graph.in_degree(node), bias, bias_constraint)
        } for node in range(N)
    }

    return BooleanNetwork.from_dict(bn_dict)


def er_boolean_network(N=10, p=0.2, bias=0.5, bias_constraint='soft', remove_multiedges=True, niter_remove=1000):
    """
    TODO: description
    """
    er_graph = nx.erdos_renyi_graph(N, p, directed=True)

    # the configuration graph creates a multigraph with self loops
    # the self loops are OK, but we should only have one copy of each edge
    if remove_multiedges:
        er_graph = _remove_duplicate_edges(graph=er_graph, niter_remove=niter_remove)

    # A dict that contains the network logic {<id>:{'name':<string>,'in':<list-input-node-id>,'out':<list-output-transitions>},..}
    bn_dict = {
        node: {
            'name': str(node),
            'in': sorted([n for n in er_graph.predecessors(node)]),
            'out': random_automata_table(er_graph.in_degree(node), bias, bias_constraint)
        } for node in range(N)
    }

    return BooleanNetwork.from_dict(bn_dict)


def random_automata_table(indegree, bias, bias_constraint='soft'):
    """
    TODO: description
    """
    if bias_constraint == 'soft':
        return [int(random.random() < bias) for b in range(2**indegree)]

    elif bias_constraint == 'hard':
        n_ones = int(bias * (2**indegree))
        output = [0] * (2**indegree - n_ones) + [1] * n_ones
        random.shuffle(output)
        return output

    elif bias_constraint == 'soft_no_constant':
        output = [int(random.random() < bias) for b in range(2**indegree)]
        if sum(output) == 0:
            output[0] = 1
            random.shuffle(output)
        return output


def _remove_duplicate_edges(graph, niter_remove=100):
    """
    TODO: description
    """
    edge_list = list(graph.edges())
    edge_frequency = Counter(edge_list)

    duplicate_edges = [edge for edge, num_edge in edge_frequency.items() if num_edge > 1]

    iremove_iter = 0
    while len(duplicate_edges) > 0 and iremove_iter < niter_remove:
        for dedge in duplicate_edges:
            if edge_frequency[dedge] > 1:
                switch_edge = random.choice(edge_list)

                # exchange edges
                graph.remove_edge(dedge[0], dedge[1])
                graph.remove_edge(switch_edge[0], switch_edge[1])
                graph.add_edge(dedge[0], switch_edge[1])
                graph.add_edge(switch_edge[0], dedge[1])

                edge_frequency[dedge] -= 1
                edge_frequency[switch_edge] -= 1

                edge_frequency[(dedge[0], switch_edge[1])] += 1
                edge_frequency[(switch_edge[0], dedge[1])] += 1

                edge_list = list(graph.edges())

        duplicate_edges = [edge for edge, num_edge in edge_frequency.items() if num_edge > 1]
        iremove_iter += 1

    if iremove_iter >= niter_remove:
        print("Warning: multi-edges were not successfully removed after %s iterations!!" % str(iremove_iter))

    return graph


def from_string_boolean(self, string, keep_constants=True, **kwargs):
    """
    Instanciates a Boolean Network from a Boolean update rules format.

    Args:
        string (string) : A boolean update rules format representation of a Boolean Network.

    Returns:
        (BooleanNetwork) : The boolean network object.

    Examples:
        String should be structured as follow:

        .. code-block:: text

            # BOOLEAN RULES (this is a comment)
            # node_name*=node_input_1 [logic operator] node_input_2 ...
            NODE3*=NODE1 AND NODE2 ...

    See also:
        :func:`from_string` :func:`from_dict`
    """

    logic = defaultdict(dict)

    # parse lines to receive node names
    network_file = StringIO(string)
    line = network_file.readline()
    i = 0
    while line != "":
        if line[0] == '#':
            line = network_file.readline()
            continue
        logic[i] = {
            'name': line.split("*")[0].strip(),
            'in': [],
            'out': []
        }
        line = network_file.readline()
        i += 1

    # Parse lines again to determine inputs and output sequence
    network_file = StringIO(string)
    line = network_file.readline()
    i = 0
    while line != "":
        if line[0] == '#':
            line = network_file.readline()
            continue
        eval_line = line.split("=")[1]  # logical condition to evaluate
        # RE checks for non-alphanumeric character before/after node name (node names are included in other node names)
        # Additional characters added to eval_line to avoid start/end of string complications
        input_names = [logic[node]['name'] for node in logic if re.compile('\W' + logic[node]['name'] + '\W').search('*' + eval_line + '*')]
        input_nums = [node for input in input_names for node in logic if input == logic[node]['name']]
        logic[i]['in'] = input_nums
        # Determine output transitions
        logic[i]['out'] = output_transitions(eval_line, input_names)
        line = network_file.readline()
        i += 1

    return self.from_dict(logic, keep_constants=keep_constants, **kwargs)



# -*- coding: utf-8 -*-
"""
Feedback Vertex Set
====================

The NP-complete Feedback Vertex Set (FVS) problem is defined as the minimum number of vertices
that need to be removed from a directed graph so that the resulting graph has no direct cycle. :cite:`Pardalos:1998` :cite:`Mochizuki:2013`.

Two methods are implemented here. A bruteforce and the Greedy Randomized Adaptive Search Procedure (GRASP) method by :cite:`Pardalos:1998`.

"""
#   Copyright (C) 2021 by
#   Alex Gates <ajgates42@gmail.com>
#   Rion Brattig Correia <rionbr@gmail.com>
#   All rights reserved.
#   MIT license.
import networkx as nx
import numpy as np
import itertools
import copy


#
# GRASP method
#
def fvs_grasp(directed_graph, max_iter=100, keep_self_loops=True):
    """The Feedback Vertex Set GRASP implementation.
    This implementation is not exact but it is recommended for very large graphs.

    Args:
        directed_graph (networkx.DiGraph) : The structure graph.
        max_iter (int) : Maximum number of iterations for the search.
        keep_self_loops (bool) : If self-loops are used in the computation. By FVS theory, all self-loop nodes are needed for control.
    Returns:
        (list) : A list of sets with the driver nodes.
    See also:
        :func:`fvs_bruteforce`.
    """
    if keep_self_loops:
        S = set(nx.nodes_with_selfloops(directed_graph))

    else:
        directed_graph = copy.deepcopy(directed_graph)
        directed_graph.remove_edges_from(directed_graph.selfloop_edges())
        S = set([])

    root_var = _root_variables(directed_graph, keep_self_loops=keep_self_loops)
    S = S.union(root_var)

    minfvc = set([frozenset(directed_graph.nodes())])

    reduced_graph = directed_graph.copy()

    for i_iter in range(max_iter):
        alpha = np.random.random()
        S = _construct_greedy_randomized_solution(reduced_graph, alpha, S)
        S = _local_search(directed_graph.copy(), S)
        compare_set = next(iter(minfvc))
        if len(S) == len(compare_set):
            minfvc.add(frozenset(S))
        elif len(S) < len(compare_set):
            minfvc = set([frozenset(S)])

    return list(minfvc)


#
# Bruteforce method
#
def fvs_bruteforce(directed_graph, max_search=5, keep_self_loops=True):
    """The Feedback Vertex Set bruteforce implementation.

    Args:
        directed_graph (networkx.DiGraph) : The structure graph.
        max_search (int) : Maximum number of additional variables to include in the search.
        keep_self_loops (bool) : If self-loops are used in the computation. By FVS theory, all self-loop nodes are needed for control.
    Returns:
        (list) : A list of sets with with the driver nodes.
    Warning:
        Use the GRASP method for large graphs.
    See also:
        :func:`fvs_grasp`.

    """
    if keep_self_loops:
        minfvc = set(nx.nodes_with_selfloops(directed_graph))

    else:
        directed_graph = copy.deepcopy(directed_graph)
        directed_graph.remove_edges_from(directed_graph.selfloop_edges())
        minfvc = set([])

    root_var = _root_variables(directed_graph, keep_self_loops=keep_self_loops)
    minfvc = minfvc.union(root_var)

    if _is_acyclic(directed_graph):
        return [minfvc]

    else:
        FVC_sets = []
        nonfvc_variables = set(directed_graph.nodes()) - minfvc

        num_additional_var = 0
        while num_additional_var <= max_search and len(FVC_sets) == 0:
            for an_combo in itertools.combinations(nonfvc_variables, num_additional_var):
                possible_fvs = minfvc.union(an_combo)

                if _is_acyclic(_graph_minus(directed_graph, possible_fvs)):
                    FVC_sets.append(possible_fvs)

            num_additional_var += 1

        return FVC_sets


def _graph_minus(graph, nodeset):
    """
    """
    newgraph = nx.DiGraph()
    for (n1, n2) in graph.edges():
        if n1 not in nodeset and n2 not in nodeset:
            newgraph.add_edge(n1, n2)

    for n in graph.nodes():
        if n not in nodeset:
            newgraph.add_node(n)

    return newgraph


def _is_acyclic(graph):
    """
    """
    return nx.is_directed_acyclic_graph(graph)


def _in0out0(directed_graph, S):
    """
    """
    remove_set = set([n for n in directed_graph.nodes()
                      if (directed_graph.in_degree(n) == 0 or directed_graph.out_degree(n) == 0)])
    directed_graph.remove_nodes_from(remove_set)
    S = S.union(remove_set)
    return directed_graph, S, len(remove_set) == 0


def _in1(directed_graph):
    """
    """
    remove_set = set([n for n in directed_graph.nodes() if (directed_graph.in_degree(n) == 1)])

    for u in remove_set:
        v = list(directed_graph.predecessors(u))[0]
        directed_graph.add_edges_from([(v, w) for w in directed_graph.successors(u)])
    directed_graph.remove_nodes_from(remove_set)
    return directed_graph, len(remove_set) == 0


def _out1(directed_graph):
    """
    """
    remove_set = set([n for n in directed_graph.nodes() if (directed_graph.out_degree(n) == 1)])

    for u in remove_set:
        v = list(directed_graph.successors(u))[0]
        directed_graph.add_edges_from([(w, v) for w in directed_graph.predecessors(u)])
    directed_graph.remove_nodes_from(remove_set)

    return directed_graph, len(remove_set) == 0


def _selfloop(directed_graph, S):
    """
    """
    remove_set = list(nx.nodes_with_selfloops(directed_graph))

    S = S.union(set(remove_set))

    directed_graph.remove_nodes_from(remove_set)
    return directed_graph, S, len(remove_set) == 0


def _reduce_instance_size(directed_graph, S):
    """
    """
    all_done = False

    while not all_done:
        all_done = True

        reduced_graph, S, no_change = _selfloop(directed_graph, S)
        all_done *= no_change

        reduced_graph, S, no_change = _in0out0(reduced_graph, S)
        all_done *= no_change

        reduced_graph, no_change = _in1(reduced_graph)
        all_done *= no_change

        reduced_graph, no_change = _out1(reduced_graph)
        all_done *= no_change

    return reduced_graph, S


def _restricted_candidate_list(directed_graph, alpha):
    """
    """
    nodelist = directed_graph.nodes()

    G = np.multiply(directed_graph.in_degree(nodelist), directed_graph.out_degree(nodelist))
    Gmin = min(G)
    Gmax = max(G)

    return np.array(nodelist)[G >= (Gmin + alpha * (Gmax - Gmin))]


def _construct_greedy_randomized_solution(directed_graph, alpha, S):
    """"
    """
    reduced_graph, S = _reduce_instance_size(directed_graph, S)

    while len(reduced_graph) > 0:
        RCL = set(_restricted_candidate_list(reduced_graph, alpha))
        s = np.random.choice(RCL)
        S.add(s)
        reduced_graph.remove_node(s)
        reduced_graph, S = _reduce_instance_size(reduced_graph, S)

    return S


def _local_search(directed_graph, S):
    """
    """
    keep_going = True

    while keep_going:
        keep_going = False
        for localnode in S:
            remove_set = S.copy()
            remove_set.discard(localnode)
            reduced_graph = directed_graph.copy()
            reduced_graph.remove_nodes_from(remove_set)
            if nx.is_directed_acyclic_graph(reduced_graph):
                S = remove_set
                keep_going = True
                break

    return S


def _root_variables(directed_graph, keep_self_loops=True):
    """
    """
    return set([n for n in directed_graph.nodes()
                if (directed_graph.in_degree(n) == 0) or ((not keep_self_loops) and (directed_graph.neighbors(n) == [n]))])



# -*- coding: utf-8 -*-
"""
Minimum Dominating Set
=======================


"""
#   Copyright (C) 2021 by
#   Alex Gates <ajgates42@gmail.com>
#   Rion Brattig Correia <rionbr@gmail.com>
#   All rights reserved.
#   MIT license.
import itertools


#
# Minimum Dominating Set
#
def mds(directed_graph, max_search=5, keep_self_loops=True):
    """The minimum dominating set method.

    Args:
        directed_graph (networkx.DiGraph) : The structural graph.
        max_search (int) : Maximum search of additional variables. Defaults to 5.
        keep_self_loops (bool) : If self-loops are used in the computation.
    Returns:
        (list) : A list of sets with the driver nodes.
    """
    N = len(directed_graph)
    root_var = _root_variables(directed_graph, keep_self_loops=keep_self_loops)

    if len(_get_dominated_set(directed_graph, root_var)) == N:
        return [root_var]
    else:
        MDS_sets = []
        nonroot_variables = set(directed_graph.nodes()) - set(root_var)
        for num_additional_var in range(1, max_search):
            for an_combo in itertools.combinations(nonroot_variables, num_additional_var):
                possible_dvs = root_var.union(an_combo)
                if len(_get_dominated_set(directed_graph, possible_dvs)) == N:
                    MDS_sets.append(possible_dvs)
            if len(MDS_sets) > 0:
                break
        return MDS_sets


def _get_dominated_set(directed_graph, dominatingset):
    """
    TODO
    """
    dominatedset = set(dominatingset)
    for dn in dominatingset:
        dominatedset.update(directed_graph.neighbors(dn))
    return dominatedset


def _root_variables(directed_graph, keep_self_loops=True):
    """
    """
    return set([n for n in directed_graph.nodes()
                if (directed_graph.in_degree(n) == 0) or ((not keep_self_loops) and (directed_graph.neighbors(n) == [n]))])



from setuptools import setup, find_packages
from cana import __package__, __description__, __version__
from Cython.Build import cythonize


def readme():
    with open('README.md') as f:
        return f.read()


# cythonize awesomeness
ext_modules = ["cana/cutils.pyx", "cana/canalization/cboolean_canalization.pyx"]


setup(
    name=__package__,
    version=__version__,
    description=__description__,
    long_description=__description__,
    long_description_content_type="text/plain",
    classifiers=[
        'Development Status :: 4 - Beta',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2.7',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Information Analysis',
    ],
    keywords="boolean networks canalization redundancy dynamical systems computational biology",
    url="http://github.com/rionbr/CANA",
    author="Alex Gates & Rion Brattig Correia",
    author_email="rionbr@gmail.com",
    license="MIT",
    packages=find_packages(),
    package_data={
        'datasets': [
            'cana.datasets/*.txt',
            'cana.datasets/bns/*.cnet',
            'cana.datasets/cell_collective/*.txt'
        ],
    },
    install_requires=[
        'numpy',
        'scipy',
        'networkx',
        'pandas',
        'cython'
    ],
    include_package_data=True,
    zip_safe=False,
    ext_modules=cythonize(ext_modules, include_path=[''], compiler_directives={'language_level': '3'})  # cython awesomeness
)



# -*- coding: utf-8 -*-
"""
Structural Controllability
===========================


"""
#   Copyright (C) 2021 by
#   Alex Gates <ajgates42@gmail.com>
#   Rion Brattig Correia <rionbr@gmail.com>
#   All rights reserved.
#   MIT license.
import networkx as nx


def sc(directed_graph, keep_self_loops=True):
    """The Structural Contolability driver sets.

    Parameters:
        directed_graph (networkx.DiGraph) : The directed graph capturing the network structure.
        keep_self_loops (bool) : Determines if the self loops are kept in the determination of structural control.

    Returns:
        (list) : A list of sets with the driver nodes.

    See also:
        If you only need the size of the set, see :func:`sc_min_size`.
    """
    # get the bipartite represenation of the directed graph
    bipartite_graph, vertex_out, vertex_in = _directed_to_bipartite(directed_graph, keep_self_loops=keep_self_loops)

    # do the maximum matching on the bipartite representation
    max_matching = nx.bipartite.hopcroft_karp_matching(bipartite_graph, top_nodes=vertex_out)

    all_max_matchings = _enumerate_maximum_matchings(bipartite_graph, vertex_out, vertex_in, max_matching)

    sc_sets = []
    for M in all_max_matchings:
        # find the vertex set whose in-coming vertices were matched
        matched_incoming = set(vertex_in) & set(M.keys())
        matched_original_vertices = set([sink[0] for sink in matched_incoming])
        new_dv_set = sorted(set(directed_graph.nodes()) - matched_original_vertices)
        if new_dv_set not in sc_sets:
            sc_sets.append(sorted(new_dv_set))

    return sc_sets


def sc_min_size(directed_graph, keep_self_loops=True):
    """The minimum number of driver variables required by structural controllability.

    Parameters:
        directed_graph (networkx.DiGraph) : The directed graph capturing the network structure.
        keep_self_loops (bool) : Determines if the self loops are kept in the determination of structural control.

    TODO:
        Implement the removal of self loops

    Returns:
        (int) : The number of driver variables necessary to render the graph structurally controlled.

    See also:
        :func:`sc`
    """
    # get the bipartite represenation of the directed graph
    bipartite_graph, vertex_out, vertex_in = _directed_to_bipartite(directed_graph, keep_self_loops=keep_self_loops)

    # do the maximum matching on the bipartite representation
    max_matching = nx.bipartite.hopcroft_karp_matching(bipartite_graph)

    # find the vertex set whose in-coming vertices were matched
    matched_incoming = set(vertex_in) & set(max_matching.keys())
    matched_original_vertices = set([sink[0] for sink in matched_incoming])

    # driver variables are those whose in-coming edges were not matched
    # although this set is not unique, only the size of the set is gaurenteed to be
    # a minimum
    sc_min_size = len(set(directed_graph.nodes()) - matched_original_vertices)

    return sc_min_size


def _directed_to_bipartite(directed_graph, keep_self_loops=True):
    """Create the undirected Bipartite representation of a directed graph.

    In this represtation, each vertex is a tuple (n, direction)
    where  one vertex set (labeled with "+") denotes the out-going
    directed edges while the other vertex set (labeled with "-") denotes the
    in-coming directed edges.  Two vertices are connected with an undirected edge
    if a directed edge exsisted in the original graph.

    Args:
        directed_graph (networkx.DiGraph) : The directed graph to be transformed.

    Returns:
        bipartite_graph (networkx.Graph) : The bipartite representation of the directed graph.
        vertex_out (list) : The list of out-going vertices.
        vertex_in (list) : The list of in-coming vertices.
    """
    # get the two vertex sets
    vertex_out = [(n, True) for n in directed_graph.nodes()]
    vertex_in = [(n, False) for n in directed_graph.nodes()]

    # make the bipartite graph
    bipartite_graph = nx.Graph()
    bipartite_graph.add_nodes_from(vertex_out, bipartite=0)
    bipartite_graph.add_nodes_from(vertex_in, bipartite=1)

    # add the edges
    for n in directed_graph.nodes():
        for v in directed_graph.successors(n):
            if keep_self_loops:
                bipartite_graph.add_edge((n, True), (v, False))
            elif (n != v):
                bipartite_graph.add_edge((n, True), (v, False))

    return bipartite_graph, vertex_out, vertex_in


def _bipartite_to_directed(bipartite_graph):
    """Recover the directed graph from its bipartite representation.

    Args:
        bipartite_graph (networkx.Graph) : The bipartite represtation of a directed graph consisting of two vertex sets:
            vertex_out and vertex_in, to be transformed back to a directed graph.

    Returns:
        directed_graph (networkx.DiGraph) : The directed graph.

    See also:
        :attr:`_directed_to_bipartite`.
    """
    directed_graph = nx.DiGraph()
    for edge, d in bipartite_graph.edges():
        for node in edge:
            if node[1] is True:
                source = node[0]
            elif node[1] is False:
                sink = node[0]
        directed_graph.add_edge(source, sink)
    return directed_graph


def _enumerate_maximum_matchings(bipartite_graph, vertex_out, vertex_in, max_matching):
    """
    @@@ Uno (1999) ???

    TODO:
        Could not find the original paper to cite here.
    """
    # this is the first matching in our list
    matchings_list = [max_matching]

    D = _make_matching_digraph(bipartite_graph, vertex_out, vertex_in, max_matching)
    D = _trim_unnecessary_edges(D)
    matchings_list = _enumerate_maximum_matchings_iter(bipartite_graph, vertex_out, vertex_in, max_matching, D, matchings_list)

    return matchings_list


def _make_matching_digraph(bipartite_graph, U, V, M):
    """
    @@@ Uno (1999) ???

    TODO:
        Could not find the original paper to cite here.

    """
    matching_digraph = nx.DiGraph()
    matching_digraph.add_nodes_from(bipartite_graph.nodes())

    matched_vertices = M.keys()

    # loop over the first vertex set
    for v in V:
        # and loop over all neighbors of v
        for u in bipartite_graph.neighbors(v):
            # check if the edge is in the maximum matching
            if v in matched_vertices and M[v] == u:
                # if the edge is matched, place a directed edge from U to V
                matching_digraph.add_edge(u, v)
            else:
                # otherwise, place a directed edge from V to U
                matching_digraph.add_edge(v, u)

    return matching_digraph


def _trim_unnecessary_edges(matching_digraph):
    """
    """
    scc_subgraphs = [matching_digraph.subgraph(scc) for scc in nx.strongly_connected_components(matching_digraph)]
    trimmed_graph = scc_subgraphs[0]
    for next_subgraph in scc_subgraphs[1:]:
        trimmed_graph = nx.union(trimmed_graph, next_subgraph)
    return trimmed_graph


def _enumerate_maximum_matchings_iter(G, U, V, M, D, matchings_list):
    """

    """
    if len(G) > 0 and not (D is None):

        # find the cycles in the matching digraph
        cycles = [c for c in nx.simple_cycles(D)]

        if len(cycles) > 0:
            # swap the edges in the cycle
            e, Mprime = _swap_edges_in_cycle(cycles[0], M)
            # this creates a new maximum matching, see if we already have it or add it
            if Mprime not in matchings_list:
                matchings_list.append(Mprime)

            # now we have two subpromblems which result in new iterations
            # Problem plus:
            Gplus, Uplus, Vplus, Mplus = _make_graph_plus(G, U, V, M, e)
            Dplus = _make_matching_digraph(Gplus, Uplus, Vplus, Mplus)
            Dplus = _trim_unnecessary_edges(Dplus)
            matchings_list = _enumerate_maximum_matchings_iter(Gplus, Uplus, Vplus, Mplus, Dplus, matchings_list)

            # and Problem minus:
            Gminus = _make_graph_minus(G, e)
            Dminus = _make_matching_digraph(Gminus, U, V, Mprime)
            Dminus = _trim_unnecessary_edges(Dminus)
            matchings_list = _enumerate_maximum_matchings_iter(Gminus, U, V, Mprime, Dminus, matchings_list)

        else:
            # if there are no cycles call the iter again without the matching digraph
            matchings_list = _enumerate_maximum_matchings_iter(G, U, V, M, None, matchings_list)

    elif len(G) > 0:
        path = _find_path_length_two(G, V, M)
        if not (path is None):
            # swap edges in the path
            e, Mprime = _swap_edges_in_path(path, M)
            # this creates a new maximum matching, see if we already have it or add it
            if Mprime not in matchings_list:
                matchings_list.append(Mprime)

            # now we have two subpromblems which result in new iterations
            # Problem plus:
            Gplus, Uplus, Vplus, Mplus = _make_graph_plus(G, U, V, M, e)
            matchings_list = _enumerate_maximum_matchings_iter(Gplus, Uplus, Vplus, Mprime, None, matchings_list)

            # Problem minus with M:
            Gminus = _make_graph_minus(G, e)
            matchings_list = _enumerate_maximum_matchings_iter(Gminus, U, V, M, None, matchings_list)

    return matchings_list


def _make_graph_plus(G, U, V, M, e):
    """
    """
    Gplus = G.copy()
    # remove the two endpoints and all adjacent edges
    Gplus.remove_node(e[0])
    Gplus.remove_node(e[1])

    Uplus = set(U)
    Vplus = set(V)
    if e[0] in U:
        Uplus.remove(e[0])
        Vplus.remove(e[1])
    else:
        Uplus.remove(e[1])
        Vplus.remove(e[0])

    Mplus = dict(M)

    return Gplus, Uplus, Vplus, Mplus


def _make_graph_minus(G, e):
    """
    """
    Gminus = G.copy()
    # remove the edge
    Gminus.remove_edge(e[0], e[1])

    return Gminus


def _find_path_length_two(G, V, M):
    """
    """
    path = None
    for v in _listminus(V, M.keys()):
        for u in G.neighbors(v):
            for w in G.neighbors(u):
                if w in M and M[w] == u:
                    path = [v, u, w]
                    break
    return path


def _listminus(list1, list2):
    """
    """
    return [a for a in list1 if a not in list2]


def _swap_edges_in_cycle(cycle, M):
    """
    """
    Mprime = dict(M)
    e = None
    cycle = cycle + [cycle[0]]
    for i in range(len(cycle) - 1):
        if cycle[i] in Mprime:
            if M[cycle[i]] == cycle[i + 1]:
                e = (cycle[i], cycle[i + 1])
                Mprime[cycle[i]] = cycle[i - 1]
                Mprime[cycle[i - 1]] = Mprime[cycle[i]]
            else:
                e = (cycle[i], cycle[i - 1])
                Mprime[cycle[i]] = cycle[i + 1]
                Mprime[cycle[i + 1]] = Mprime[cycle[i]]
    return e, Mprime


def _swap_edges_in_path(path, M):
    """
    """
    Mprime = dict(M)
    e = (path[2], path[1])
    m = Mprime[path[2]]
    del Mprime[path[2]]
    del Mprime[m]
    Mprime[path[0]] = path[1]
    Mprime[path[1]] = path[0]
    return e, Mprime
